"""
This module defines functions that calculate various orientational quantities.
"""

import numpy as np
from math import cos,sin,atan2,sqrt,radians
               
def right_handed_coord_system(x,y,z):
    """
    Making a set of 3 vectors (x,y,z) a right-handed coordinate system
    Parameters:
    x, y, z: column array vectors representing the principal axes
    """
    det_matrix = np.linalg.det(np.concatenate((x,y,z),axis=1))
    if det_matrix < 0:
        x = -x
    else:
        x = x
    return(x,y,z)

def quat_to_mat(quat):
    """
    Return rotation matrix from quaternion
    quat is an array of 4 elements
    """
    w2 = quat[0]*quat[0]
    i2 = quat[1]*quat[1]
    j2 = quat[2]*quat[2]
    k2 = quat[3]*quat[3]
    twoij = 2.0*quat[1]*quat[2]
    twoik = 2.0*quat[1]*quat[3]
    twojk = 2.0*quat[2]*quat[3]
    twoiw = 2.0*quat[1]*quat[0]
    twojw = 2.0*quat[2]*quat[0]
    twokw = 2.0*quat[3]*quat[0]

    m1 = []
    m1.extend([w2+i2-j2-k2])
    m1.extend([twoij-twokw])
    m1.extend([twojw+twoik])
    m1 = np.asarray(m1)

    m2 = []
    m2.extend([twoij+twokw])
    m2.extend([w2-i2+j2-k2])
    m2.extend([twojk-twoiw])
    m2 = np.asarray(m2)
    
    m3 = []
    m3.extend([twoik-twojw])
    m3.extend([twojk+twoiw])
    m3.extend([w2-i2-j2+k2])
    m3 = np.asarray(m3)

    return (m1,m2,m3)

def compute_quaternion_from_principal_axes(x,y,z):
    """
    Parameters:
    x, y, z: column array vectors representing the principal axes
    
    Returns: quaternions
    quat_w: scalar part of the quaternion
    quat_i: quaternion along the i axis of the quaternion vector part
    quat_j: quaternion along the j axis of the quaternion vector part
    quat_k: quaternion along the k axis of the quaternion vector part  
    
    Modified from: https://github.com/lammps/lammps
    math_extra.cpp
    Copyright (2003) Sandia Corporation.
    """
    # Right-handed coordinate system:
    x,y,z = right_handed_coord_system(x,y,z)

    # Squares of quaternion components:
    q0sq = 0.25 * (x[0] + y[1] + z[2] + 1.0)
    q1sq = q0sq - 0.5 * (y[1] + z[2])
    q2sq = q0sq - 0.5 * (x[0] + z[2])
    q3sq = q0sq - 0.5 * (x[0] + y[1])
    # Some component must be greater than 1/4 since they sum to 1
    # compute other compomemts from it
    q = np.empty(4, dtype=float) 
    if q0sq >= 0.25:
        q[0] = sqrt(q0sq)
        q[1] = (y[2] - z[1]) / (4.0*q[0])
        q[2] = (z[0] - x[2]) / (4.0*q[0])
        q[3] = (x[1] - y[0]) / (4.0*q[0])
    elif q1sq >= 0.25:
        q[1] = sqrt(q1sq)
        q[0] = (y[2] - z[1]) / (4.0*q[1])
        q[2] = (y[0] + x[1]) / (4.0*q[1])
        q[3] = (x[2] + z[0]) / (4.0*q[1])
    elif q2sq >= 0.25:
        q[2] = sqrt(q2sq)
        q[0] = (z[0] - x[2]) / (4.0*q[2])
        q[1] = (y[0] + x[1]) / (4.0*q[2])
        q[3] = (z[1] + y[2]) / (4.0*q[2])
    elif q3sq >= 0.25:
        q[3] = sqrt(q3sq)
        q[0] = (x[1] - y[0]) / (4.0*q[3])
        q[1] = (z[0] + x[2]) / (4.0*q[3])
        q[2] = (z[1] + y[2]) / (4.0*q[3])
    norm = 1/sqrt(q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3])
    q[0] = norm*q[0]
    q[1] = norm*q[1]
    q[2] = norm*q[2]
    q[3] = norm*q[3]
    return (q) 

def compute_principal_axes_from_q(q):
    """
    Computing principal axes from quaternions
    ############################
    Parameters:
    q = [quat_w,quat_i,quat_j,quat_k]
    where
    quat_w: scalar part of the quaternion
    quat_i: quaternion along the i axis of the quaternion vector part
    quat_j: quaternion along the j axis of the quaternion vector part
    quat_k: quaternion along the k axis of the quaternion vector part

    #############################

    Returns: 
    x, y, z: column array vectors representing the principal axes
    
    Modified from: https://github.com/lammps/lammps
    math_extra.cpp
    Copyright (2003) Sandia Corporation.
    """
    x,y,z = np.empty(3, dtype=float)

    x[0] = q[0]*q[0] + q[1]*q[1] - q[2]*q[2] - q[3]*q[3]
    x[1] = 2.0 * (q[1]*q[2] + q[0]*q[3])
    x[2] = 2.0 * (q[1]*q[3] - q[0]*q[2])

    y[0] = 2.0 * (q[1]*q[2] - q[0]*q[3])
    y[1] = q[0]*q[0] - q[1]*q[1] + q[2]*q[2] - q[3]*q[3]
    y[2] = 2.0 * (q[2]*q[3] + q[0]*q[1])

    z[0] = 2.0 * (q[1]*q[3] + q[0]*q[2])
    z[1] = 2.0 * (q[2]*q[3] - q[0]*q[1])
    z[2] = q[0]*q[0] - q[1]*q[1] - q[2]*q[2] + q[3]*q[3]

    return(x,y,z)

def eulerAnglesToRotationMatrix(theta) :
    """
    Input:
    Theta = an array with 3 element corresponding to 3 Euler angles (degrees). 
    
    Return:
    R_z1,R_x2,Rz3 = arrays of vectors that define the rotation
    R = Rotation matrix
    """
    theta = [radians(i) for i in theta]
    R_z1 = np.array([[cos(theta[0]),    -sin(theta[0]),    0],
                    [sin(theta[0]),    cos(theta[0]),     0],
                    [0,                     0,                      1]
                    ])
                           
    R_x2 = np.array([[1,         0,                  0                   ],
                    [0,         cos(theta[1]), -sin(theta[1]) ],
                    [0,         sin(theta[1]), cos(theta[1])  ]
                    ])
                 
    R_z3 = np.array([[cos(theta[2]),    -sin(theta[2]),    0],
                    [sin(theta[2]),    cos(theta[2]),     0],
                    [0,                     0,                      1]
                    ])
                        
    R = np.dot(R_z1, np.dot( R_x2, R_z3 ))
     
    return(R_z1,R_x2,R_z3,R)

def isRotationMatrix(R) :
    Rt = np.transpose(R)
    shouldBeIdentity = np.dot(Rt, R)
    I = np.identity(3, dtype = R.dtype)
    n = np.linalg.norm(I - shouldBeIdentity)
    return n < 1e-6
 
def rotationMatrixToEulerAngles(R) :
    """
    Returns Euler Angles from Rotation matrix, in radians
    Rotation Matrix: numpy matrix with columns as body-fixed principal axes
    """
    assert(isRotationMatrix(R))
     
    sy = sqrt(R[0,2] * R[0,2] +  R[1,2] * R[1,2])
     
    singular = sy < 1e-6
 
    if  not singular :
        x = atan2(R[0,2] , -R[1,2])
        y = atan2(sy,R[2,2])
        z = atan2(R[2,0], R[2,1])
    else:
        x = atan2(R[1,0], R[0,0])/2
        y = atan2(sy,R[2,2])
        z = atan2(R[1,0], R[0,0])/2
    return [x, y, z]



"""
This module defines general-purpose functions.
"""

from collections import Counter, OrderedDict
import numpy as np

def clamp(n, minn, maxn):
    "Restrict number n in a range [minn,maxn]"
    return max(min(maxn, n), minn)

class OrderedCounter(Counter, OrderedDict):
    'Counter that remembers the order elements are first encountered'

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, OrderedDict(self))

    def __reduce__(self):
        return self.__class__, (OrderedDict(self),)

def countLineWithPattern(file, num_molecule):
    mol_ID = [("site"+str(i)+"_") for i in np.linspace(1,num_molecule,num_molecule,dtype=int)] 
    results = OrderedCounter(mol_ID)
    myfile = open(file)
    for line in myfile:
        for pattern in mol_ID:
            if pattern in line:
                results[pattern] +=1
    # For some reasons counter returns one extra count
    results = np.array([(i[1]-1) for i in list(results.items())])
    return results

def frame_split(a, chunk):
    n = -(-len(a)//chunk)
    k, m = divmod(len(a), n)
    return list(a[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))

def get_coeff(path,coeff_file):
    """
    Get coefficients from coefficient file
    Return: (m,1) matrix where m is the number of basis function coefficients
    """
    coeff = []
    with open(path + coeff_file,"r") as infile:
        for line in infile:
            data = [float(i) for i in line.rstrip('\t\n').split("\t")]
            coeff.append(data)
    coeff = np.matrix(coeff)
    return coeff



"""
This module defines functions that calculate molecular coordinates.
"""

import numpy as np
import sys

def min_image(dx,xbox,xbox2):
    """
    Calculates minimum image separations in one dimension x using 
    coordinates of lower and upper limits of the simulation box,
    xboxlo and xboxhi, and box width and half width, xbox and xbox2.
    
    Returns dx
    """
    while (dx > xbox2):
        dx = dx - xbox
    while (dx < -xbox2):
        dx = dx + xbox
    return(dx)

def compute_unwrapped_coords(x,y,z,xboxlo,xboxhi,yboxlo,yboxhi,zboxlo,zboxhi):
    """
    Takes lists of x,y,z coordinates of sites (for which periodic boundary
    conditions may have been applied) that are ordered such that adjacent
    sites in the list should be separated by less than half the simulation
    box lenght and returns their unwrapped (absolute) coords measured position
    relative to the coords of the 1st site in the list
    
    Parameters:
    x,y,z: lists of possibly wrapped coords in x, y and z 
    xboxlo,xboxhi,yboxlo,yboxhi,zboxlo,zboxhi: box boundaries
    
    Returns:
    xu,yu,zu: unwrapped coords in x, y and z
    """
    
    if (len(x) == len(y)) and (len(x) == len(z)):
        nSites = len(x)
    else:
        print("Error: coord arrays in compute_unwrapped_coords unequal in length", file=sys.stderr)
        sys.exit(1)
    
    xbox = xboxhi - xboxlo
    xbox2 = 0.5*xbox
    ybox = yboxhi - yboxlo
    ybox2 = 0.5*ybox
    zbox = zboxhi - zboxlo
    zbox2 = 0.5*zbox
    
    xu = [e for e in x]
    yu = [e for e in y]
    zu = [e for e in z]
    
    xi = xu[0]
    yi = yu[0]
    zi = zu[0]
    
    for j in range(nSites):
        dxji = xu[j] - xi
        dyji = yu[j] - yi
        dzji = zu[j] - zi
        dxji = min_image(dxji,xbox,xbox2)
        dyji = min_image(dyji,ybox,ybox2)
        dzji = min_image(dzji,zbox,zbox2)
        xu[j] = round(xi + dxji,3)
        yu[j] = round(yi + dyji,3)
        zu[j] = round(zi + dzji,3)
        
        xi = xu[j]
        yi = yu[j]
        zi = zu[j]
    # Make sure most sites are inside simulation box:
    x_centre = np.mean(xu)
    y_centre = np.mean(yu)
    z_centre = np.mean(zu)

    if x_centre < xboxlo:
        for i in xu:
            i = i + xbox
    elif x_centre > xboxhi:
        for i in xu:
            i = i - xbox
    if y_centre < yboxlo:
        for i in yu:
            i = i + ybox
    elif y_centre > yboxhi:
        for i in yu:
            i = i - ybox
    if z_centre < zboxlo:
        for i in zu:
            i = i + zbox
    elif z_centre > zboxhi:
        for i in zu:
            i = i - zbox
    return(xu,yu,zu)

def right_handed_coord_system(e_x,e_y,e_z):
    """
    Make a set of 3 vectors assuming to be (x,y,z) arranged right-handed 
    Parameters: 
    e_x, e_y, e_z: column array vectors representing the principal axes of the body
    Returns:
    e_x, e_y, e_z: column array vectors representing the principal axes of the body arranged right-handed
    """
    det_matrix = np.linalg.det(np.concatenate((e_x,e_y,e_z),axis=1))
    if det_matrix < 0:
        e_x = -e_x
    else:
        e_x = e_x
    return(e_x,e_y,e_z)


def global_to_body_transform(R,e_x,e_y,e_z):
    """
    Transformation between global and body frame for a set of 3 vectors 
    that represent principal axes of a CG site. Body frame here is the
    reference CG site
    Parameters:
    R: transformation matrix from global to body frame
    e_x,e_y,e_z: column array vectors representing the principal axes of a CG site
    Returns:
    e_x,e_y,e_z: column array vectors representing the principal axes of a CG site
    in the reference CG site body frame
    """
    e_x = np.dot(R,e_x)
    e_y = np.dot(R,e_y)
    e_z = np.dot(R,e_z)
    return np.concatenate((e_x,e_y,e_z), axis=1)
    
def global_to_body_frame(R,e_x,e_y,e_z):
    """
    Transformation between global and body frame for a set of 3 vectors 
    that represent principal axes of a CG site. Body frame here is the
    reference CG site
    Parameters:
    R: transformation matrix from global to body frame
    e_x,e_y,e_z: column array vectors representing the principal axes of a CG site
    Returns:
    e_x,e_y,e_z: column array vectors representing the principal axes of a CG site
    in the reference CG site body frame
    """
    e_x = np.dot(R,e_x)
    e_y = np.dot(R,e_y)
    e_z = np.dot(R,e_z)
    return (e_x,e_y,e_z)

def global_to_body_vector(R,t,r):
    """
    Transformation between global and body frame for a vector
    This is useful to calculate (x,y,z) coordinate of
    a CG site wrt the reference CG site (to then calculate spherical
    coordinate of that CG site wrt the ref CG site)
    Parameters:
    R: transformation matrix from global to body frame
    t: translation column vector from global to body frame
    r: column array vector representing the coordinate of a CG site in the lab frame
    Returns:
    r: new column array vector representing the coordinate of a CG site
    in the reference body frame
    """
    r = np.dot(R,r+t)
    return r

def pbc_image(xboxlo,xboxhi,yboxlo,yboxhi,zboxlo,zboxhi,site_pos,cutoff):
    """
    Output periodic images. Maximum 3/6 images outputted depending on 
    the site position. This is made to cover non-bonded interactions
    that are caused by pbc  images of the molecules in the simulation box.
    Parameters: xboxlo,xboxhi,yboxlo,yboxhi,zboxlo,zboxhi are box boundaries,
                type int
                site_pos is site position [x,y,z] type "list" 
    Returns: list of positions of images
    """
    #cutoff = 15 # have to add this parameters in later
    xToLo = abs(site_pos[0] - xboxlo)
    xToHi = abs(site_pos[0] - xboxhi)
    yToLo = abs(site_pos[1] - yboxlo)
    yToHi = abs(site_pos[1] - yboxhi)
    zToLo = abs(site_pos[2] - zboxlo)
    zToHi = abs(site_pos[2] - zboxhi)
    site_pos_box = [xToLo,xToHi,yToLo,yToHi,zToLo,zToHi]
    image = []
    num_image = sum(i < cutoff for i in site_pos_box)
    if num_image == 3:
        if xToLo < cutoff:
            if yToLo < cutoff:
                if zToLo < cutoff:
                    image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
                elif zToHi < cutoff:
                    image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
            elif yToHi < cutoff:
                if zToLo < cutoff:
                    image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
                elif zToHi < cutoff:
                    image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
        elif xToHi < cutoff:
            if yToLo < cutoff:
                if zToLo < cutoff:
                    image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
                elif zToHi < cutoff:
                    image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
            elif yToHi < cutoff:
                if zToLo < cutoff:
                    image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
                elif zToHi < cutoff:
                    image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
    elif num_image == 2:
        if xToLo >= cutoff and xToHi >= cutoff:
            if yToLo < cutoff:
                if zToLo < cutoff:
                    image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
                elif zToHi < cutoff:
                    image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
            elif yToHi < cutoff:
                if zToLo < cutoff:
                    image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
                elif zToHi < cutoff:
                    image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
        elif yToLo >= cutoff and yToHi >= cutoff:
            if xToLo < cutoff:
                if zToLo < cutoff:
                    image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
                elif zToHi < cutoff:
                    image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
            elif xToHi < cutoff:
                if zToLo < cutoff:
                    image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
                elif zToHi < cutoff:
                    image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
        elif zToLo >= cutoff and zToHi >= cutoff:
            if yToLo < cutoff:
                if xToLo < cutoff:
                    image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
                elif xToHi < cutoff:
                    image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
            elif yToHi < cutoff:
                if xToLo < cutoff:
                    image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
                elif xToHi < cutoff:
                    image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
                    image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
    elif num_image == 1:
        if xToLo < cutoff:
            image.append([site_pos[0]+xboxhi-xboxlo,site_pos[1],site_pos[2]])
        elif xToHi < cutoff:
            image.append([site_pos[0]-(xboxhi-xboxlo),site_pos[1],site_pos[2]])
        elif yToLo < cutoff:
            image.append([site_pos[0],site_pos[1]+yboxhi-yboxlo,site_pos[2]])
        elif yToHi < cutoff:
            image.append([site_pos[0],site_pos[1]-(yboxhi-yboxlo),site_pos[2]])
        elif zToLo < cutoff:
            image.append([site_pos[0],site_pos[1],site_pos[2]+zboxhi-zboxlo])
        elif zToHi < cutoff:
            image.append([site_pos[0],site_pos[1],site_pos[2]-(zboxhi-zboxlo)])
    elif num_image == 0:
        image = [0]
    else:
        print("ERROR: Minimum image condition is not apply. \
              More than 3 pbc images are found")
        sys.exit(1)
    return image



"""
This module defines the a function that calculate the matrix block for 1 atomistic frame (3 rows of the whole matrix)
"""

import numpy as np
from math import sqrt,acos

from .pbf import grid_centre, radius_norm_1D, periodic_basis_force_torque_cubic_splines,grid_centre_nonuniform
from .read_lammps_data import read_atomistic_data, output_CG_data, box_boundaries_Lammps_file
from .coords_functions import compute_unwrapped_coords, global_to_body_transform, global_to_body_frame,right_handed_coord_system, min_image
from .orientation_functions import rotationMatrixToEulerAngles
from .helper_functions import clamp

def construct_matrix(frame,path,file_name,r_min,r_max,num_r_center,k_max,num_molecule,num_atom_per_molecule,spheroid):
    """
    Contruct force matrix for a frame (a system configuration) from molecular 
    dynamic trajectory.
    The matrix should have (3*num_molecule) rows corresponding to the number of 
    atomistic force data points, and (num_r_center*k_max**3+1) columns
    corresponding to the number of basis functions+1. 
    The last column is the atomistic force on a CG particle.
    """
    cutoff = r_max
    center_r = grid_centre_nonuniform(r_min,r_max,num_r_center)

    # Create a file to save the matrix into hard disk
    read_atomistic_data(path,file_name,num_atom_per_molecule,num_molecule,frame)
    force_AA, torque_AA = output_CG_data(path,file_name,frame,num_atom_per_molecule,num_molecule,spheroid)
    matrix_f, matrix_t = [], []

    # Create files to save the matrix into hard disk
    matrix_file_f = open(path  + "scratch/" + file_name + "_matrix_force_" + str(frame),'w')
    matrix_file_f.close()
    matrix_file_t = open(path  + "scratch/" + file_name + "_matrix_torque_" + str(frame),'w')
    matrix_file_t.close()

    # Simulation box
    xboxlo,xboxhi,yboxlo,yboxhi,zboxlo,zboxhi = box_boundaries_Lammps_file(path,file_name,num_atom_per_molecule,num_molecule,frame)
    xbox = xboxhi - xboxlo
    xbox2 = 0.5*xbox
    ybox = yboxhi - yboxlo
    ybox2 = 0.5*ybox
    zbox = zboxhi - zboxlo
    zbox2 = 0.5*zbox
    
    # Calculate matrix components
    with open (path + "scratch/" + file_name + "_frame_" + str(frame) + "_CG_data",'r') as fin, open(path  + "scratch/" + file_name + "_matrix_force_" + str(frame),'a') as fout, open(path  + "scratch/" + file_name + "_matrix_torque_" + str(frame),'a') as tout:
        CG_coord = fin.readlines()
        for i in range(num_molecule):
            pbf_all_f,pbf_all_t = [], []
            matrix_xyz_f, matrix_xyz_t = [[] for _ in range(3)],[[] for _ in range(3)]
            
            ref_site_pos = np.array([float(CG_coord[i].split()[1]),float(CG_coord[i].split()[2]),float(CG_coord[i].split()[3])])
            ref_site_orient_x = np.reshape([float(CG_coord[i].split()[4]),float(CG_coord[i].split()[5]),float(CG_coord[i].split()[6])],(3,1))
            ref_site_orient_y = np.reshape([float(CG_coord[i].split()[7]),float(CG_coord[i].split()[8]),float(CG_coord[i].split()[9])],(3,1))
            ref_site_orient_z = np.reshape([float(CG_coord[i].split()[10]),float(CG_coord[i].split()[11]),float(CG_coord[i].split()[12])],(3,1))
            ref_site_orient_x, ref_site_orient_y, ref_site_orient_z=right_handed_coord_system(ref_site_orient_x, ref_site_orient_y, ref_site_orient_z)
            ref_site_orient_matrix = np.concatenate((np.transpose(ref_site_orient_x),np.transpose(ref_site_orient_y),np.transpose(ref_site_orient_z)),axis=0)
            neighbour_list = list(range(num_molecule))
            neighbour_list.remove(i)
            for j in neighbour_list:
                site_pos = np.array([float(CG_coord[j].split()[1]),
                        float(CG_coord[j].split()[2]),float(CG_coord[j].split()[3])])
                dx= min_image(ref_site_pos[0]-site_pos[0],xbox,xbox2)
                dy= min_image(ref_site_pos[1]-site_pos[1],xbox,xbox2)
                dz= min_image(ref_site_pos[2]-site_pos[2],xbox,xbox2)
                site_site_dist = sqrt(dx**2 + dy**2 + dz**2)
                if site_site_dist<0.1 or site_site_dist > cutoff:
                        pass
                else:
                    site_orient_x = np.reshape([float(CG_coord[j].split()[4]),float(CG_coord[j].split()[5]),float(CG_coord[j].split()[6])],(3,1))
                    site_orient_y = np.reshape([float(CG_coord[j].split()[7]),float(CG_coord[j].split()[8]),float(CG_coord[j].split()[9])],(3,1))
                    site_orient_z = np.reshape([float(CG_coord[j].split()[10]),float(CG_coord[j].split()[11]),float(CG_coord[j].split()[12])],(3,1))
                    site_orient_x, site_orient_y, site_orient_z=right_handed_coord_system(site_orient_x, site_orient_y, site_orient_z)
                    site_orient_matrix_glob = np.concatenate((np.transpose(ref_site_orient_x),np.transpose(ref_site_orient_y),np.transpose(ref_site_orient_z)),axis=0)
                    site_orient_matrix_body = global_to_body_transform(ref_site_orient_matrix,site_orient_x,site_orient_y,site_orient_z)
                    beta = rotationMatrixToEulerAngles(site_orient_matrix_body)[1]
                    # recalculate site_pos due to periodic boundaries
                    site_pos = np.array([ref_site_pos[0]- dx,ref_site_pos[1]- dy,ref_site_pos[2]- dz])
                    site_site_dir = (ref_site_pos - site_pos)/np.linalg.norm(ref_site_pos - site_pos)
                    A = acos(clamp(float(np.dot(site_site_dir,ref_site_orient_z)),-1,1))
                    B = acos(clamp(float(np.dot(site_site_dir,site_orient_z)),-1,1))
                    dcosA_dR = (ref_site_orient_z.flatten() - site_site_dir)*float(np.dot(site_site_dir,ref_site_orient_z))/site_site_dist
                    dcosB_dR = (site_orient_z.flatten() - site_site_dir)*float(np.dot(site_site_dir,site_orient_z))/site_site_dist
                    dcosbeta_di3 = site_orient_z.flatten() 
                    dcosA_di3 = site_site_dir
                    i3_unit = ref_site_orient_z.flatten() 
                    basis_var = [site_site_dist,beta,A,B]

                    pbf_f, pbf_t =  periodic_basis_force_torque_cubic_splines(center_r,k_max,basis_var,site_site_dir,dcosA_dR,dcosB_dR,dcosbeta_di3,dcosA_di3,i3_unit)
                    pbf_all_f.append(pbf_f)
                    pbf_all_t.append(pbf_t)
            #Sum for all neighbour sites
            pbf_all_f=[sum(i) for i in zip(*pbf_all_f)]
            pbf_all_t=[sum(i) for i in zip(*pbf_all_t)]
            for pbf_i in range(len(pbf_all_f)):
                for k in range(3):
                    matrix_xyz_f[k].append(pbf_all_f[pbf_i][k])
                    matrix_xyz_t[k].append(pbf_all_t[pbf_i][k])
            matrix_f.extend(matrix_xyz_f)
            matrix_t.extend(matrix_xyz_t)

        matrix_f = np.column_stack([np.array(matrix_f),np.array(force_AA)])
        matrix_t = np.column_stack([np.array(matrix_t),np.array(torque_AA)])
        np.savetxt(fout, matrix_f, fmt="%.6e")
        np.savetxt(tout, matrix_t, fmt="%.6e")
    return 



from .coords_functions import *
from .helper_functions import *
from .matrix import *
from .orientation_functions import *
from .pbf import *
from .read_lammps_data import *
from .solver_SEQHT import *



"""
This module defines functions to read and process LAMMPS dump trajectory files.
"""

from itertools import islice
import numpy as np
from math import acos

from .coords_functions import compute_unwrapped_coords

def column_name_Lammps_file(path,file_name):
    """
    Ouput columns of interested information such as: coordinates x,y,z, mass, forces
    fx, fy, fz, charge.
    """
    with open(path + file_name, "r") as input:
        for line in islice(input,8,9):
            col_var = line.split()
            mol_id = [i for i,x in enumerate(col_var) if x== "mol"][0] - 2
            x_col = [i for i,x in enumerate(col_var) if x== "x"][0] - 2
            y_col = [i for i,x in enumerate(col_var) if x== "y"][0] - 2
            z_col = [i for i,x in enumerate(col_var) if x== "z"][0] - 2
            fx_col = [i for i,x in enumerate(col_var) if x== "fx"][0] - 2
            fy_col = [i for i,x in enumerate(col_var) if x== "fy"][0] - 2
            fz_col = [i for i,x in enumerate(col_var) if x== "fz"][0] - 2
            mass_col = [i for i,x in enumerate(col_var) if x== "mass"][0] - 2
    return (mol_id,x_col,y_col,z_col,fx_col,fy_col,fz_col,mass_col)

def read_atomistic_data(path,file_name,num_atom_per_molecule,num_molecule,frame):
    """
    Read LAMMPS dump file and output coordinates and forces information

    Parameters:
    x_col,y_col,z_col,mass_col,fx_col,fy_col,fz_col = column containing
    x,y,z,mass,fx,fy,fz in LAMMPS dump file
    frame = which frame to be read

    Returns:
    A file containing x_col,y_col,z_col,fx_col,fy_col,fz_col,mass_col
    """
    # Lines to read from LAMMPS dump file
    first_line = (frame-1)*(num_atom_per_molecule*num_molecule+9)+9
    last_line = frame*(num_atom_per_molecule*num_molecule+9)-1

    # Write out files of interested time frames:
    mol_id,x_col,y_col,z_col,fx_col,fy_col,fz_col,mass_col = column_name_Lammps_file(path,file_name)
    with open(path + file_name, "r") as input, open(path + "scratch/" + file_name + "_frame_" + str(frame),"w") as output:
        for line in islice(input, first_line, last_line+1):
            data = line.split()[mol_id],line.split()[x_col],line.split()[y_col],line.split()[z_col],line.split()[fx_col],\
            line.split()[fy_col],line.split()[fz_col],line.split()[mass_col]
            output.write("\t".join(data) + "\n")
    with open (path + "scratch/" + file_name + "_frame_" + str(frame),"r+") as input:
        data = input.readlines()
        data.sort(key=lambda l: float(l.split("\t")[0]),reverse=False)
        input.seek(0)
        input.truncate()
        input.write("".join(data))
    return

def box_boundaries_Lammps_file(path,file_name,num_atom_per_molecule,num_molecule,frame):
    """
    Output boundaries of simulation box set for a simulation with dump file
    "path/file_name"
    """
    boundary_x_line = (frame-1)*(num_atom_per_molecule*num_molecule+9)+5

    with open(path + file_name, "r") as lammps_file:
        for line in islice(lammps_file, boundary_x_line, boundary_x_line+1):
            xboxlo = float(line.split(" ")[0])
            xboxhi = float(line.split(" ")[1].rstrip('\n'))
        for line in islice(lammps_file, 0, 1):
            yboxlo = float(line.split(" ")[0])
            yboxhi = float(line.split(" ")[1].rstrip('\n'))
        for line in islice(lammps_file, 0, 1):
            zboxlo = float(line.split(" ")[0])
            zboxhi = float(line.split(" ")[1].rstrip('\n'))
    return (xboxlo,xboxhi,yboxlo,yboxhi,zboxlo,zboxhi)

def output_CG_data(path,file_name,frame,num_atom_per_molecule,num_molecule,spheroid):
    """
    Output CG data out for each time frame.
    """
    xyz=[]
    m=[]
    site_ID = 1
    # Box boundaries:
    xboxlo,xboxhi,yboxlo,yboxhi,zboxlo,zboxhi = box_boundaries_Lammps_file(path,file_name,num_atom_per_molecule,num_molecule,frame)
    # Read file:
    force_AA = []
    torque_AA = []
    with open (path + "scratch/" + file_name + "_frame_" + str(frame),'r') as input, open(path + "scratch/" + file_name + "_frame_" + str(frame) + "_CG_data","w") as output:
        while True:
            try:
                data = [next(input) for _ in range(num_atom_per_molecule)]
                x = [float(i.split()[1]) for i in data]
                y = [float(i.split()[2]) for i in data]
                z = [float(i.split()[3]) for i in data]
                fx = [float(i.split()[4]) for i in data]
                fy = [float(i.split()[5]) for i in data]
                fz = [float(i.split()[6]) for i in data]
                m = [float(i.split()[7]) for i in data]

                # Unwrapping atom coords under periodic boundary conditions:
                x,y,z = compute_unwrapped_coords(x,y,z,xboxlo,xboxhi,yboxlo,yboxhi,zboxlo,zboxhi)
                # Calculating the centre of mass
                center = np.array([np.dot(x,m)/sum(m),np.dot(y,m)/sum(m),np.dot(z,m)/sum(m)])
                # Calculating vectors connecting center of mass and each atoms:
                x = np.reshape(x, (num_atom_per_molecule,1))
                y = np.reshape(y, (num_atom_per_molecule,1))
                z = np.reshape(z, (num_atom_per_molecule,1))
                m_matrix = np.diag(m)
                xyz = np.concatenate((x,y,z), axis=1)
                coord = np.matrix(xyz) - center

                # Calculating moment of inertia:
                weighted_coord = m_matrix*coord
                inertia_dot = np.array(np.dot(coord.transpose(), weighted_coord))
                inertia = -inertia_dot
                inertia[0][0] = inertia_dot[1][1] + inertia_dot[2][2]
                inertia[1][1] = inertia_dot[0][0] + inertia_dot[2][2]
                inertia[2][2] = inertia_dot[0][0] + inertia_dot[1][1]

                e_values, e_vectors = np.linalg.eig(inertia)

                order = np.argsort(e_values)
                eval1, eval2, eval3 = e_values[order]
                if spheroid == "oblate":
                    z_prime = np.squeeze(np.asarray(e_vectors[:,order[2]]))
                    y_prime = np.squeeze(np.asarray(e_vectors[:,order[1]]))
                    x_prime = np.squeeze(np.asarray(e_vectors[:,order[0]]))
                elif spheroid == "prolate":
                    z_prime = np.squeeze(np.asarray(e_vectors[:,order[0]]))
                    y_prime = np.squeeze(np.asarray(e_vectors[:,order[1]]))
                    x_prime = np.squeeze(np.asarray(e_vectors[:,order[2]]))
                output.write("{} {} {} {} {} {} {} {} {} {} {} {} {} \n".format(
                    site_ID,center[0],center[1],center[2],x_prime[0],x_prime[1],
                    x_prime[2],y_prime[0],y_prime[1],y_prime[2],z_prime[0],
                    z_prime[1],z_prime[2]))

                force_AA.extend([sum(fx),sum(fy),sum(fz)])
                torque = np.sum([np.cross(coord[i],np.array([fx[i],fy[i],fz[i]])) for i in range(len(fx))],axis=0).tolist()[0]
                torque_AA.extend([torque[0],torque[1],torque[2]])
                site_ID +=1
                
            except StopIteration:
                break
    return(force_AA,torque_AA)



import os
import shutil
import numpy as np
import h5py

def read_dense_matrix_in_block_from_h5_file(path,file_name,file_ext,block_index,block_size_full):
    """
    Load the LQ matrix from a matrix data file in h5 compressed format.
    """
    with h5py.File(path + file_name + "_matrix_" + file_ext + "_gzip.h5",'r') as hf:
        data_all = hf.get('matrix')
        data = data_all[block_size_full*(block_index-1):block_size_full*block_index,:]
    return data

def SEQHT(path,file_name,file_ext,num_frames,block_size,num_molecule,W):
    """
    Sequential householder transformation to reduce a matrix to a square matrix
    which has the same least-squares solution and residual.
    return: 
    W: squared matrix
    resid: force-torque residue 
    """    
    block_size_full = int(3*num_molecule*block_size)
    q = int(3*num_molecule*num_frames/block_size)
    for t in range(1,q+1):
        At = read_dense_matrix_in_block_from_h5_file(path,file_name,file_ext,t,block_size_full)
        W = np.concatenate((W,At),axis=0)
        # Series of Householder transformation to make W become triangular form, similar to QR decomposition
        W = np.linalg.qr(W)[1]
    return (W)

def LS_solver(path,file_name,file_ext,W):
    """
    Solve a matrix using linear least-squares method and write out solution file
    Return the residual of the least-squares solution
    """
    x = np.linalg.lstsq(W[:,0:-1],W[:,-1],rcond=None)[0]
    with open(path + file_name + "_coeff_frame_"+file_ext,"w") as outfile:
        for j in x:
            outfile.write(str(float(j)) + "\t")
            outfile.write("\n")
    resid = abs(W[-1][-1])
    #Remove path/scratch directory:
    #shutil.rmtree(path + "scratch")
    return (resid)



import setuptools

setuptools.setup(
    name="afmcg",
    version="1.0.1",
    license="CC-BY 4.0",
    author="Huong TL Nguyen",
    author_email="huong.nguyen@adelaide.edu.au",
    description="Anisotropic coarse-graining of atomistic dynamical simulations using force-matching",
    url="https://github.com/dmhuanglab/afmcg",
    download_url = "https://github.com/dmhuanglab/afmcg/archive/refs/tags/v1.0.1.tar.gz",
    keywords = ['coarse-grained','parametrization','force-matching'],
    packages=setuptools.find_packages(),
    install_requires=(
        'matplotlib',
        'numpy',
        'h5py'
    ),
    classifiers=[
        "License :: OSI Approved",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
)



"""
This module contains functions to calculate basis functions
"""

import numpy as np
from math import cos, sin, pi

def grid_centre(var_min,var_max,grid):
    """
    Compute centres for grids of data range in term of individual variables
    """
    num_centre = int((var_max-var_min)/grid)
    centre = np.linspace(var_min+grid/2,var_max-grid/2,num_centre)
    return(centre)

def grid_centre_nonuniform(var_min,var_max,N):
    """
    Compute centres for grids of data range in term of individual variables
    N is number of centers
    """
    centre = []
    for i in range(0,N):
        centre.append(var_max-(var_max-var_min)*cos(i*pi/(2*(N-1)))) 
    return(centre)

def pbf_meshgrid(center_r):
    """
    """
    k1 = np.arange(0,9,2)
    k2 = np.arange(0,9,2)
    k3 = np.arange(0,9,2)
    pbf_meshgrid = np.array(np.meshgrid(center_r,k1,k2,k3)).T.reshape(-1,4)
    return pbf_meshgrid
    
def radius_norm_1D(centre,x,grid):
    """
    Computing the norm of distance from a point vector(x)
    about a point vector(centre) in 1D
    """
    d_norm = ((x-centre)*2/grid)
    return(d_norm)

def cosine_der(k,x):
    """
    x is cos(y)
    evaluate derivative of cos(ky) wrt x=cos(y) for k = 0,1,2,3,4,...
    """
    if sin(x) == 0 or k==0:
        cos_der = 0
    else:
        cos_der = (k*sin(k*x))/sin(x)
    return cos_der
    
def nat_cubic_spline_basis(x,x_c,x_c_all):
    """
    Calculate natural cubic spline basis functions
    """
    ppart = lambda t: np.maximum(0, t)
    cube = lambda t: t*t*t
    numerator = (cube(ppart(x - x_c)) - cube(ppart(x - x_c_all[-1])))
    denominator = x_c_all[-1] - x_c
    return (numerator / denominator)

def nat_cubic_spline_basis_dR(x,x_c,x_c_all):
    """
    Calulate first derivative of natural cubic spline basis functions
    """
    ppart = lambda t: np.maximum(0, t)
    square = lambda t: t*t
    numerator = (3*square(ppart(x - x_c)) - 3*square(ppart(x - x_c_all[-1])))
    denominator = x_c_all[-1] - x_c
    return (numerator / denominator)

def periodic_basis_cubic_splines(center_r,k_max,data):
    """
    Evaluate periodic basis function (pbf) for CG intermolecular potential. Output is a set of basis function calculated for each neighbor of a CG site
    
    Parameters:
    pbf_i: 4D array containing important indices to construct basis function
    data: 4D array (r, beta, A, B)
    R_unit: 3D unit vector
    A_d_R: 3D vector
    B_d_R: 3D vector
    grid_r

    Returns:
    pbf: float value
    """

    pbf = []
    r,beta,A,B = data[:]

    # Periodic basis function terms
    for i in range(-2,len(center_r)-2):
        cub_spl_fixed = nat_cubic_spline_basis(r,center_r[-2],center_r)
        if i==-2:
            cub_spl = 1
        elif i==-1:
            cub_spl = r
        else:
            cub_spl = nat_cubic_spline_basis(r,center_r[i],center_r) - cub_spl_fixed
        for k1 in range(0,k_max+1,2):
            gk1 = cos(k1*beta)
            for k2 in np.arange(0,k_max+1,2):
                gk2 = cos(k2*A)
                for k3 in np.arange(0,k_max+1,2):
                    gk3 = cos(k3*B)
                    pbf.append(cub_spl*gk1*gk2*gk3)
    return pbf

def periodic_basis_force_torque_cubic_splines(center_r,k_max,basis_var,R_unit,dcosA_dR,dcosB_dR,dcosbeta_di3,dcosA_di3,i3_unit):
    """
    Evaluate periodic basis function (pbf) first erivative wrt CG site-site vectors (pbf_dR),
    for basis functions for intermolecular forces. Output is a set of basis function calculated
    for each neighbor of a CG site

    Parameters:
    pbf_i: 4D array containing important indices to construct basis function
    data: 4D array (r, beta, A, B)
    R_unit: 3D unit vector
    A_d_R: 3D vector
    B_d_R: 3D vector
    grid_r

    Returns:
    pbf_dR: 3D vector
    pbf_di3: 3D vector
    """
    r,beta,A,B = basis_var[:]
    pbf_dR,pbf_di3 = [], []

    # Periodic basis function terms
    for i in range(-2,len(center_r)-2):
        cub_spl_fixed = nat_cubic_spline_basis(r,center_r[-2],center_r)
        cub_spl_fixed_dR = nat_cubic_spline_basis_dR(r,center_r[-2],center_r)*R_unit
        if i==-2:
            cub_spl = 1
            cub_spl_dR = np.array([0,0,0])
        elif i==-1:
            cub_spl = r
            cub_spl_dR = R_unit
        else:
            cub_spl = nat_cubic_spline_basis(r,center_r[i],center_r) - cub_spl_fixed
            cub_spl_dR = nat_cubic_spline_basis_dR(r,center_r[i],center_r)*R_unit - cub_spl_fixed_dR
        for k1 in range(0,k_max+1,2):
            gk1 = cos(k1*beta)
            dgk1_dR = np.array([0,0,0])
            dgk1_di3 = cosine_der(k1,beta)*dcosbeta_di3
            for k2 in np.arange(0,k_max+1,2):
                gk2 = cos(k2*A)
                dgk2_dR = cosine_der(k2,A)*dcosA_dR
                dgk2_di3 = cosine_der(k2,A)*dcosA_di3
                for k3 in np.arange(0,k_max+1,2):
                    gk3 = cos(k3*B)
                    dgk3_dR = cosine_der(k3,B)*dcosB_dR
                    # The minus sign in front of each pbf_dR element is due to F= -dU/dR
                    pbf_dR.append(-(cub_spl_dR*gk1*gk2*gk3 + dgk1_dR*cub_spl*gk2*gk3
                            + dgk2_dR*cub_spl*gk1*gk3 + dgk3_dR*cub_spl*gk1*gk2))
                    pbf_di3.append(-np.cross(i3_unit,(dgk1_di3*cub_spl*gk2*gk3
                            + dgk2_di3*cub_spl*gk1*gk3)))
    return(pbf_dR,pbf_di3)

def pbf_matrix(r_min,r_max,num_r_center,k_max,data):
    """
    Compute pbf matrix for a data list that contains elementary lists such as
    [x,beta,phi,theta] for x,beta,phi,theta are the separation distance and
    angles that define relative orientations between two uniaxial particles
    
    center_r: list of knots for basis functions of inter-particle distance
    k_max: maximum order of periodic basis functions
    data: list of lists [[r_1,beta_1,phi_1,theta_1],...,[r_n,beta_n,phi_n,theta_n]]

    Return:
    matrix array with size (len(data),len(center_r)*(k_max/2+1)**3 
    """
    # Centre points for site-site distance variable:
    center_r = grid_centre_nonuniform(r_min,r_max,num_r_center)

    matrix_array = []
    matrix_append = matrix_array.append
    for i in data:
        pbf = periodic_basis_cubic_splines(center_r,k_max,i)
        matrix_append(pbf)
    matrix_array = np.matrix(matrix_array)
    return matrix_array
